@@ -1,6 +1,7 @@
 """Primary backend for Zeno. Handles all data processing and caching."""
 
 import asyncio
+import glob
 import logging
 import os
 import pickle
@@ -22,7 +23,7 @@ from zeno.api import (
     ZenoOptions,
     ZenoParameters,
 )
-from zeno.classes.base import MetadataType, ZenoColumnType
+from zeno.classes.base import DataProcessingReturn, MetadataType, ZenoColumnType
 from zeno.classes.classes import MetricKey, TableRequest, ZenoColumn
 from zeno.classes.report import Report
 from zeno.classes.slice import FilterIds, FilterPredicateGroup, Slice, SliceMetric
@@ -228,6 +229,21 @@ class ZenoBackend(object):
         self.status = "Done processing"
         print(self.status)
 
+    def __set_data_processing_returns(self, rets: List[List[DataProcessingReturn]]):
+        """Update DataFrame with new columns from processing functions.
+
+        Args:
+            rets (List[List[DataProcessingReturn]]): List of returns from decorated
+            functions.
+        """
+        for ret in rets:
+            for out in ret:
+                c_hash = str(out.column)
+                self.df.loc[:, c_hash] = out.output
+                self.df[c_hash] = self.df[c_hash].convert_dtypes()
+                out.column.metadata_type = getMetadataType(self.df[c_hash])
+                self.complete_columns.append(out.column)
+
     def __predistill(self) -> None:
         """Run distilling functions not dependent on model outputs."""
 
@@ -239,8 +255,8 @@ class ZenoBackend(object):
             save_path = Path(self.cache_path, str(predistill_column) + ".pickle")
 
             load_series(self.df, predistill_column, save_path)
-            predistill_hash = str(predistill_column)
 
+            predistill_hash = str(predistill_column)
             if self.df[predistill_hash].isnull().any():
                 predistill_to_run.append(predistill_column)
             else:
@@ -262,11 +278,7 @@ class ZenoBackend(object):
                     [self.batch_size] * len(predistill_to_run),
                     range(len(predistill_to_run)),
                 )
-                for out in predistill_outputs:
-                    self.df.loc[:, str(out[0])] = out[1]
-                    self.df[str(out[0])] = self.df[str(out[0])].convert_dtypes()
-                    out[0].metadata_type = getMetadataType(self.df[str(out[0])])
-                    self.complete_columns.append(out[0])
+                self.__set_data_processing_returns(predistill_outputs)
 
     def __inference(self):
         """Run models on instances."""
@@ -302,6 +314,24 @@ class ZenoBackend(object):
                 model_column.metadata_type = getMetadataType(self.df[model_hash])
                 self.complete_columns.append(model_column)
 
+                # Check if there were saved postdistill columns:
+                for f in glob.glob(
+                    os.path.join(
+                        self.cache_path, "POSTDISTILL*" + model_name + ".pickle"
+                    )
+                ):
+                    name = os.path.basename(f).split(model_name)[0][11:]
+                    col = ZenoColumn(
+                        column_type=ZenoColumnType.POSTDISTILL,
+                        name=name,
+                        model=model_name,
+                    )
+                    series = pd.read_pickle(f)
+                    self.df.loc[:, str(col)] = series
+                    self.df[str(col)] = self.df[str(col)].convert_dtypes()
+                    col.metadata_type = getMetadataType(self.df[str(col)])
+                    self.complete_columns.append(col)
+
         if len(models_to_run) > 0:
             with Pool() as pool:
                 inference_outputs = pool.map(
@@ -314,14 +344,7 @@ class ZenoBackend(object):
                     [self.batch_size] * len(models_to_run),
                     range(len(models_to_run)),
                 )
-                for out in inference_outputs:
-                    self.df.loc[:, str(out[0])] = out[2]
-                    self.df[str(out[0])] = self.df[str(out[0])].convert_dtypes()
-                    # If we get an embedding, add it to DataFrame.
-                    if not out[3].isnull().values.any():  # type: ignore
-                        self.df.loc[:, str(out[1])] = out[3]
-                    out[0].metadata_type = getMetadataType(self.df[str(out[0])])
-                    self.complete_columns.append(out[0])
+                self.__set_data_processing_returns(inference_outputs)
 
     def __postdistill(self) -> None:
         """Run distill functions dependent on model outputs."""
@@ -337,6 +360,11 @@ class ZenoBackend(object):
                 }
             )
             col_hash = str(col_name)
+
+            # If we already loaded in inference, skip.
+            if col_hash in self.df.columns:
+                continue
+
             save_path = Path(self.cache_path, col_hash + ".pickle")
 
             load_series(self.df, col_name, save_path)
@@ -360,11 +388,7 @@ class ZenoBackend(object):
                     [self.batch_size] * len(postdistill_to_run),
                     range(len(postdistill_to_run)),
                 )
-                for out in post_outputs:
-                    self.df.loc[:, str(out[0])] = out[1]  # type: ignore
-                    self.df[str(out[0])] = self.df[str(out[0])].convert_dtypes()
-                    out[0].metadata_type = getMetadataType(self.df[str(out[0])])
-                    self.complete_columns.append(out[0])
+            self.__set_data_processing_returns(post_outputs)
 
     def get_metrics_for_slices(
         self,
